#!/usr/bin/perl
use strict;
use warnings;
use feature 'state';
use Fcntl qw(:flock);
use File::Spec;
use Getopt::Long qw(:config bundling no_ignore_case pass_through);
use POSIX qw(ceil);

my %opts = (
	quiet => 0,
	recurse => 0,
	tabstop => 4,
);

GetOptions(
	\%opts,
	'tabstop=i',
	'recurse!',
	'quiet',
) or die "Can't parse options!\n";

die "Nothing to do!\n" unless @ARGV;

sub whitewash {
	my $s = shift;
	$s =~ s/^\s+//g;
	$s =~ s/\s+$//g;
	$s =~ s/\s+/ /g;
	return $s;
}

my $sect_rgx = '\[\S+\]';
my $var_rgx = '^(\$?[\w\.-]+) = ';

sub clean_file {
	my $file = shift;
	open(my $fh, '<', $file) or die "Can't open $file: $!\n";
	local $/ = "\r\n";
	my @file = <$fh>;
	close $fh;
	my @tab_counts;
	my $word_len;
	my $l = 1;
	foreach(@file) {
		s/\s*=\s*/ = /;
		s/\s*,/, /g;
		s/\s*;\s*/ ; /g;
		s/(?:\s*;\s*)+$//; #Remove empty comments
		$_ = whitewash $_;

		if($_ =~ /$sect_rgx/o) {
			push @tab_counts, $word_len if defined $word_len;
			$word_len = 0;
		} elsif($_ =~ /$var_rgx/o) {
			my $var = $1;
			my $len = length $var;
			no warnings;
			$word_len = $len if $len > $word_len;
		}
		$l++;
	}

	defined $word_len ? push @tab_counts, $word_len : return;
	my $sect_total = @tab_counts;
	$_ = int($_ / $opts{tabstop}) + 1 foreach @tab_counts;

	$l = 1;
	foreach(@file) {
		state $tab_count;
		state $cf;
		if(/$sect_rgx/o) {
			die "I am a bug; we counted $sect_total sections and ran out at line $l\n" unless @tab_counts;
			$tab_count = shift @tab_counts;
		} elsif(/$var_rgx/o) {
			die "Fail: tab count should not be $tab_count; $file:$l\n" unless $tab_count;
			my $t = "\t" x ceil(($opts{tabstop} * $tab_count - length($1)) / $opts{tabstop});
			s/ = /$t=\t/;
			$cf = 1 if /,$/;
		} elsif($_ && $cf) {
			my $pad = "\t" x ($tab_count + 1);
			$_ = $pad.$_;
		} elsif(!$_) {
			undef $cf;
		}
		$l++;
	}
	s/\s+$// foreach @file;
	$_ .= "\r\n" foreach @file;
	die "I am a bug; we ended up with @tab_counts extra sections after parsing $file\n" if @tab_counts;
	open($fh, '>', $file) or die "Can't open $file: $!\n";
	flock($fh, LOCK_EX);
	print $fh (@file);
	close $fh;
}

sub process_inputs {
	foreach(@_) {
		if(-f $_) {
			clean_file $_ if $_ =~ /\.ltx$/;
		} elsif($opts{recurse} && -d $_) {
			my $d = $_;
			$d =~ s#/$##; #Vim has some issues with non-slash regex delimiters.
			opendir(my $dh, $d) or die "Can't open $d: $!\n";
			process_inputs(map {File::Spec->catfile($d, $_)} grep !/^\./, readdir $dh);
			closedir $dh;
		} else {
			warn "I don't know what to do with $_. Getting drunk instead.\n" unless $opts{quiet};
		}
	}
}

process_inputs @ARGV;

